static void 
assign_mult_eoprec_wd_eo(const float_gpu &mkappa2, 
                               cu_wqf_eo *dye, 
                               cu_wqf_eo *ye, 
                            const cu_gvf *u, 
                           const cu_clvf *fclinv = NULL) {
//
// Multiply even/odd-site preconditioned Dirac operator : yd = D y,
// where D = 1 - kappa^2 M M
//
//  mkappa2 : hopping parameter (= -kappa^2)
//      dye : even/odd site fermion (output)
//       ye : even/odd site fermion (input)
//        u : gauge link
//   fclinv : inverse clover term
//

  cu_wqf_eo *Myo;
  int ieo = (*ye).ieo;
  int ioe = 1-ieo;

//#pragma omp parallel for
//  for (int igx=0;igx<_NGX;igx++) {
//    CUDA_SAFE_CALL(cudaThreadSynchronize());
//  }

  copy_boundary(ye);

  cutStartTimer(mult_timer);
  Myo = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*Myo).g[igx]),sizeof(cu_wqf_eo_blk)));

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,       (*ye).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    //
    //  Myo = Moe ye, or  Foo \ Moe ye
    //
    if (is_clover) {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                         (*ye).g[igx],
                                                  (*u).eo[ioe].g[igx], 
                                                  (*u).eo[ieo].g[igx],
                                             (*fclinv).eo[ioe].g[igx]);
    } else {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                         (*ye).g[igx],
                                                  (*u).eo[ioe].g[igx],
                                                  (*u).eo[ieo].g[igx]);
    }
    CUT_CHECK_ERROR("cumult_wd_in_kernel (all) fails.");
//    CUDA_SAFE_CALL(cudaThreadSynchronize());

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

  }
  cutStopTimer(mult_timer);

  copy_boundary(Myo);

  cutStartTimer(mult_timer);
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,      (*Myo).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    //
    //  Mye = Meo yo, or  Fee \ Meo yo
    //
    if (is_clover) {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*dye).g[igx],phys_blk_d.g[igx],
                                                        (*Myo).g[igx],
                                                  (*u).eo[ieo].g[igx],
                                                  (*u).eo[ioe].g[igx],(*fclinv).eo[ieo].g[igx]);
    } else {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*dye).g[igx],phys_blk_d.g[igx],
                                                        (*Myo).g[igx],
                                                  (*u).eo[ieo].g[igx],
                                                  (*u).eo[ioe].g[igx]);
    }
    CUT_CHECK_ERROR("cumult_wd_in_kernel (2all) fails.");
//    CUDA_SAFE_CALL(cudaThreadSynchronize());

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

    CUDA_SAFE_CALL(cudaFree((*Myo).g[igx]));
  }
  _mm_free(Myo);
  cutStopTimer(mult_timer);
  mult_flop_count += 2*MULT_WD;
  _multiter_ += 2;

  // dye = - kappa^2 dye + ye
  accum_mult_add(dye, mkappa2, ye);

}
