#undef _CHECKSUM_

#define STRIDE_TZ                        (sizeof(cu_wqf_eo_thd)*_NBZ)
#define NGHOST_X(_depth_)                (sizeof(cu_wqf_eo_thd)*_NBZ*_NBY*(_depth_))
#define NGHOST_Y(_depth_)                (sizeof(cu_wqf_eo_thd)*_NBZ*(_NBX+(_depth_)*2)*_NGX*(_depth_))
#define BUFF_X_IDX(iby,ibx,igx,_depth_)  (((iby-1)+(ibx+igx*(_depth_))*_NBY)*STRIDE_TZ)
#define BUFF_Y_IDX(iby,ibx,igx,_depth_)  ((iby+((ibx-1+(_depth_))+igx*(_NBX+2*(_depth_)))*(_depth_))*STRIDE_TZ)

static int gpu_is_initialized = 0;

//
// Buffer for ghost sites exchange. (allocated by cudaHostAlloc on pinned memory)
//
// GPU<=>CPU BUFF
static char *gbuff_x_up;
static char *gbuff_x_dn;
static char *gbuff_y_up;
static char *gbuff_y_dn;

// MPI BUFF
#if _NDIMX != 1
static char *sbuff_x_up;
static char *sbuff_x_dn;
static char *rbuff_x_up;
static char *rbuff_x_dn;
#endif
#if _NDIMY != 1
static char *sbuff_y_up;
static char *sbuff_y_dn;
static char *rbuff_y_up;
static char *rbuff_y_dn;
#endif

static long long int __call_counter = 0;

static inline long long int
__get_checksum(const int length, char *vec) {
  long long int csum;
  csum = 0;
  for (int i = 0; i < length; ++i) {
    csum = csum + (long long int)(vec[i]);
  }
  return csum;
}

static inline void
__sendrecv_char(const int size, char *sbuff, const int dst_rank,
                                char *rbuff, const int src_rank) {
#ifdef _QMX_
  qmx_sendrecv_char(size,sbuff,dst_rank,rbuff,src_rank);
#else
  MPI_Status sstat;
  MPI_Sendrecv((void *)sbuff,size,MPI_BYTE,dst_rank,_nodeid_     ,
               (void *)rbuff,size,MPI_BYTE,src_rank,src_rank,MPI_COMM_WORLD,&sstat);
#endif
#if defined(_CHECKSUM_)
  long long int ssum = __get_checksum(size,sbuff);
  long long int rsum = __get_checksum(size,rbuff);
  printf("%10lld: size:%12d send:%2d -> %2d %12lld  recv:%d -> %d %12lld\n",
     __call_counter,size,_nodeid_,dst_rank,ssum,src_rank,_nodeid_,rsum);
  __call_counter++;
#endif
}

static void
new_copy_boundary() {
//
// create ghost site buffers
//

  if (0 == gpu_is_initialized) {
    //
    // allocates memory for 1st call.
    //

    //
    // Pinned memory on host side
    //
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_up,NGHOST_X(_NDEPTH)*_NGX,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_x_dn,NGHOST_X(_NDEPTH)*_NGX,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_up,NGHOST_Y(_NDEPTH)     ,cudaHostAllocPortable));
    CUDA_SAFE_CALL(cudaHostAlloc((void **)&gbuff_y_dn,NGHOST_Y(_NDEPTH)     ,cudaHostAllocPortable));
    memset(gbuff_x_up,0,NGHOST_X(_NDEPTH)*_NGX);
    memset(gbuff_x_dn,0,NGHOST_X(_NDEPTH)*_NGX);
    memset(gbuff_y_up,0,NGHOST_Y(_NDEPTH));
    memset(gbuff_y_dn,0,NGHOST_Y(_NDEPTH));

    //
    // MPI buffers
    //
#if _NDIMX != 1
    sbuff_x_up = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    sbuff_x_dn = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    rbuff_x_up = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    rbuff_x_dn = (char *)_mm_malloc(NGHOST_X(_NDEPTH),16);
    memset(sbuff_x_up,0,NGHOST_X(_NDEPTH));
    memset(sbuff_x_dn,0,NGHOST_X(_NDEPTH));
    memset(rbuff_x_up,0,NGHOST_X(_NDEPTH));
    memset(rbuff_x_dn,0,NGHOST_X(_NDEPTH));
#endif
#if _NDIMY != 1
    sbuff_y_up = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    sbuff_y_dn = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    rbuff_y_up = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    rbuff_y_dn = (char *)_mm_malloc(NGHOST_Y(_NDEPTH),16);
    memset(sbuff_y_up,0,NGHOST_X(_NDEPTH));
    memset(sbuff_y_dn,0,NGHOST_X(_NDEPTH));
    memset(rbuff_y_up,0,NGHOST_X(_NDEPTH));
    memset(rbuff_y_dn,0,NGHOST_X(_NDEPTH));
#endif

    gpu_is_initialized = 1;
  }
}

static void
delete_copy_boundary() {
//
// delete ghost site buffers
//

/*
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_x_up));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_x_dn));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_y_up));
  CUDA_SAFE_CALL(cudaFreeHost(gbuff_y_dn));

#if _NDIMX != 1
  _mm_free(sbuff_x_up);
  _mm_free(sbuff_x_dn);
  _mm_free(rbuff_x_up);
  _mm_free(rbuff_x_dn);
#endif
#if _NDIMY != 1
  _mm_free(sbuff_y_up);
  _mm_free(sbuff_y_dn);
  _mm_free(rbuff_y_up);
  _mm_free(rbuff_y_dn);
#endif
*/

}

static void 
copy_boundary_send_x(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (GPU=>HOST=>GPU send phase in a node)
//

  // GPU => CPU => GPU (step 1) (within a node)
#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,iby,_NBX-depth+1+ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
      CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx,depth)]),
                                &((*ye).g[igx]->blk[blk_idx(1,iby,           1+ibx)]),
                                STRIDE_TZ,cudaMemcpyDeviceToHost));
    }
    }
  }
#pragma omp parallel for
  for (int igx = 0; igx < _NGX; igx++) {
    int igx_up = (igx+1+_NGX) % _NGX;
    int igx_dn = (igx-1+_NGX) % _NGX;
    for (int ibx = 0; ibx <= depth-1; ibx++) {
    for (int iby = 1; iby <= _NBY; iby++) {
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,     1-depth+ibx)]),
                                &(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx_dn,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
      CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),
                                &(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                                STRIDE_TZ,cudaMemcpyHostToDevice));
    }
    }
  }

}

static void
copy_boundary_sendrecv_x(const int depth){
//
// copy boundary in x direction (HOST<=>HOST MPI sendrecv phase)
//

#if _NDIMX != 1

#pragma omp parallel sections
{
#pragma omp section
  memcpy((void *)&(sbuff_x_up[0]),(void *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX-1)]),NGHOST_X(depth));
#pragma omp section
  memcpy((void *)&(sbuff_x_dn[0]),(void *)&(gbuff_x_dn[                       0]),NGHOST_X(depth));
}

  __sendrecv_char(NGHOST_X(depth),sbuff_x_up,_nodeidup_(0),rbuff_x_up,_nodeiddn_(0));
  __sendrecv_char(NGHOST_X(depth),sbuff_x_dn,_nodeiddn_(0),rbuff_x_dn,_nodeidup_(0));

#pragma omp parallel sections
{
#pragma omp section
  memcpy((void *)&(gbuff_x_up[NGHOST_X(depth)*(_NGX-1)]),(void *)&(rbuff_x_up[0]),NGHOST_X(depth));
#pragma omp section
  memcpy((void *)&(gbuff_x_dn[                       0]),(void *)&(rbuff_x_dn[0]),NGHOST_X(depth));
}

#endif
}

static void
copy_boundary_recv_x(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (HOST=>GPU recv phase)
//

#if _NDIMX != 1
  // CPU => GPU (step 4)
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    int igx_up = (igx+1+_NGX) % _NGX;
    int igx_dn = (igx-1+_NGX) % _NGX;
    if (0==igx) {
      for (int ibx = 0; ibx <= depth-1; ibx++) {
      for (int iby = 1; iby <= _NBY; iby++) {
        CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,     1-depth+ibx)]),
                                  &(gbuff_x_up[BUFF_X_IDX(iby,ibx,igx_dn,depth)]),
                                  STRIDE_TZ,cudaMemcpyHostToDevice));
      }
      }
    }
    if (_NGX-1==igx) {
      for (int ibx = 0; ibx <= depth-1; ibx++) {
      for (int iby = 1; iby <= _NBY; iby++) {
        CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,iby,      _NBX+1+ibx)]),
                                  &(gbuff_x_dn[BUFF_X_IDX(iby,ibx,igx_up,depth)]),
                                  STRIDE_TZ,cudaMemcpyHostToDevice));
      }
      }
    }
  }
#endif
  ycopy_dsize += NGHOST_X(depth)*2;
}


static void
copy_boundary_send_y(const int depth, cu_wqf_eo *ye){
//
// copy boundary in y direction (GPU=>HOST=>GPU send phase in a node)
//

#pragma omp parallel for
  for (int igx = 0; igx < _NGX;igx++) {
  for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
  for (int iby = 0; iby <= depth-1;iby++) {
    CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                              &((*ye).g[igx]->blk[blk_idx(1,_NBY+1-depth+iby,ibx)]),
                              STRIDE_TZ,cudaMemcpyDeviceToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                              &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                              STRIDE_TZ,cudaMemcpyDeviceToDevice));
  }
  }
  }

#if _NDIMY != 1

// GPU => GPU (to another GPU) (3 steps)
  // GPU => CPU (step 1)
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
  for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
  for (int iby = 0; iby <= depth-1;iby++) {
    CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                              &((*ye).g[igx]->blk[blk_idx(1,_NBY-depth+1+iby,ibx)]),
                              STRIDE_TZ,cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaMemcpy(&(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                              &((*ye).g[igx]->blk[blk_idx(1,           1+iby,ibx)]),
                              STRIDE_TZ,cudaMemcpyDeviceToHost));
  }
  }
  }
#endif
}

static void
copy_boundary_sendrecv_y(const int depth){
//
// copy boundary in y direction (HOST<=>HOST MPI sendrecv phase)
//

#if _NDIMY != 1
#pragma omp parallel sections
{
#pragma omp section
  memcpy((void *)sbuff_y_up,(void *)gbuff_y_up,NGHOST_Y(depth));
#pragma omp section
  memcpy((void *)sbuff_y_dn,(void *)gbuff_y_dn,NGHOST_Y(depth));
}

  __sendrecv_char(NGHOST_Y(depth),sbuff_y_up,_nodeidup_(1),rbuff_y_up,_nodeiddn_(1));
  __sendrecv_char(NGHOST_Y(depth),sbuff_y_dn,_nodeiddn_(1),rbuff_y_dn,_nodeidup_(1));

#pragma omp parallel sections
{
#pragma omp section
  memcpy((void *)gbuff_y_up,(void *)rbuff_y_up,NGHOST_Y(depth));
#pragma omp section
  memcpy((void *)gbuff_y_dn,(void *)rbuff_y_dn,NGHOST_Y(depth));
}
#endif
}

static void
copy_boundary_recv_y(const int depth, cu_wqf_eo *ye){
//
// copy boundary in x direction (HOST=>GPU recv phase)
//

#if _NDIMY != 1
  // CPU => GPU (step 3)
#pragma omp parallel for
  for (int igx = 0;igx < _NGX;igx++) {
  for (int ibx = 1-depth; ibx <= _NBX+depth;ibx++) {
  for (int iby = 0; iby <= depth-1;iby++) {
    CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,     1-depth+iby,ibx)]),
                              &(gbuff_y_up[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                              STRIDE_TZ,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(&((*ye).g[igx]->blk[blk_idx(1,      _NBY+1+iby,ibx)]),
                              &(gbuff_y_dn[BUFF_Y_IDX(iby,ibx,igx,depth)]),
                              STRIDE_TZ,cudaMemcpyHostToDevice));
  }
  }
  }
#endif
  ycopy_dsize += NGHOST_Y(depth)*2;
}

static void
copy_boundary_all(cu_wqf_eo *ye){
//
// copy boundary incuding extended sites
//

  cutStartTimer(ycopy_timer);

#ifndef _singlePU
#ifdef _QMX_
  qmx_barrier();
#else
  MPI_Barrier(MPI_COMM_WORLD);
#endif
#endif

  copy_boundary_send_x(_NDEPTH,ye);
  copy_boundary_sendrecv_x(_NDEPTH);
  copy_boundary_recv_x(_NDEPTH,ye);

  copy_boundary_send_y(_NDEPTH,ye);
  copy_boundary_sendrecv_y(_NDEPTH);
  copy_boundary_recv_y(_NDEPTH,ye);

  ycopy_count += 1;
  cutStopTimer(ycopy_timer);
}

static void
copy_boundary(cu_wqf_eo *ye){
//
// copy boundary for boundary edge
//

  cutStartTimer(ycopy_timer);

#ifndef _singlePU
#ifdef _QMX_
  qmx_barrier();
#else
  MPI_Barrier(MPI_COMM_WORLD);
#endif
#endif

  copy_boundary_send_x(1,ye);
  copy_boundary_sendrecv_x(1);
  copy_boundary_recv_x(1,ye);

  copy_boundary_send_y(1,ye);
  copy_boundary_sendrecv_y(1);
  copy_boundary_recv_y(1,ye);

  ycopy_count += 1;
  cutStopTimer(ycopy_timer);
}

#undef STRIDE_TZ
#undef NGHOST_X
#undef NGHOST_Y
#undef BUFF_X_IDX
#undef BUFF_Y_IDX
