#ifndef _GPU_SOLVER_CONFIG_H
#define _GPU_SOLVER_CONFIG_H

!==========================================
! User should properly define:
! _NGX, _NTHZ, _CEIL, _NRAS, _NMR,
!  and _NDEPTH
!==========================================

!******************************************
! RAS preconditioner parameters
!  NRAS : RAS iteration counts
!   NMR : block solver iteration count
!******************************************
#define _NRAS (2)
#define _NMR  (8)

!******************************************
! RAS overlap depth 
! (should be >= 1)
!******************************************
#define  _NDEPTH  (1)

!******************************************
! _NGX: Numbuer of GPU par node (X direction only)
! This should satisfy:
!
!  mod((_NTX/_NDIMX),_NGX) == 0
!
!******************************************
#define  _NGX   (1)

!******************************************
! _NTHZ: Number of threads in Z direction
! This should stasisfy:
!
!  mod((_NTZ/_NDIMZ),_NTHZ) == 0
!
!  _NTHZ*_NTHT == _NTHREADS <= Hardware limitation for total threads in a block.
!
!******************************************
#define  _NTHZ  (8)

!******************************************
! CEIL: a thread number for reduction process on a GPU.
! This number should satisfy:
!
! _NTHREADS <= _CEIL <= _NTHREADS*2
!
! _CEIL == 2^?  (a power of 2 number)
!
! _CEIL <= the max number for threads in a block (hardware limit)
!
!******************************************
#define _CEIL (256)
!#define _CEIL (128)
!#define _CEIL (64)

!==================================================
! do not edit the following lines
!==================================================
!******************************************
! threads in T direction
!******************************************
#define  _NTHT  (_NTT/2)

!******************************************
! blocks in Z-Y-X volume/GPU
!******************************************
#define  _NBX  (_NX/_NGX)
#define  _NBY   _NY
#define  _NBZ  (_NZ/_NTHZ)

#define  _NTHREADS  (_NTHZ*_NTHT)
#define  _NBLOCKS   (_NBX*_NBY*_NBZ)

#endif
