!-------------------------------------------------------------------------------
!= PHMC_HMC Configuration file, Fortran 9X, Free Format
!
!== Version
!  $Id: config.h,v 1.7 2011/09/30 10:46:08 ishikawa Exp $
!-------------------------------------------------------------------------------
#ifndef LQFT_CONFIG_H_
#define LQFT_CONFIG_H_

!-------------------------------------------
! VERSION
!-------------------------------------------
#define _VERSION_ "refactored-v1.597-SF0.942"

!-------------------------------------------
! Section: Set Lattice size
!-------------------------------------------
#define _NTX 32
#define _NTY 32
#define _NTZ 32
#define _NTT 64
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: Set node size
!-------------------------------------------
#define _NDIMX  2
#define _NDIMY  2
#define _NDIMZ  1
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: Lattice boundary condtion
!-------------------------------------------
!===========================================================================
! Schroedinger Functional Boundary setups
! 
! If _SF is defined, the SF boundary condition is employed.
! _SF_BOUNDARY_TYPE == 0, phi=pi*( -1, 0, 1)/3, phi_prime=pi*( -3, 1, 2)/3
! _SF_BOUNDARY_TYPE == 1, phi=pi*( -1, 0, 1)/6, phi_prime=pi*( -5, 2, 3)/6
!---------------------------------------------------------------------------
!#define _SF
#undef _SF
#define _SF_BOUNDARY_TYPE 0
!#define _SF_BOUNDARY_TYPE 1
!--------------
! End Section
!--------------

!-------------------------------------------
! Action configuration
!-------------------------------------------
!----------------------------
! Section: Gauge action
!----------------------------
!-------------------------------------------
! If _RGGAUGE is defined, we use gauge 
!     action including rectangle loops.
!-------------------------------------------
#define _RGGAUGE
!#undef _RGGAUGE
!--------------
! End Section
!--------------

!------------------------------------
! Section: Wilson type quark action
!------------------------------------
!-------------------------------------------
! If _CLOVER is defined, we use 
! the O(a)-improved Wilson (clover) action.
!-------------------------------------------
#define _CLOVER
!#undef _CLOVER

!-------------------------------------------
! choose hmc Wilson-Dirac matrix inverse solver
! bicgstab_alg   : BiCGStab
!-------------------------------------------
#define _HMC_SOLVER_ bicgstab_alg

!-------------------------------------------
! choose clover term matrix inverse solver
! clvinv_lu  : use LU decomposition
! clvinv_ldl : use LDL^+ decomposition
!-------------------------------------------
!#define _CLV_SOLVER  clvinv_lu
#define _CLV_SOLVER  clvinv_ldl

!--------------
! End Section
!--------------

!------------------------------------
! Section: Measurement solver
!------------------------------------
!-------------------------------------------
! measurement wilson/clover solver
!  BLOCK solver for singele NODE only
!-------------------------------------------
!#define _USE_BLOCK_SOLVER
#undef _USE_BLOCK_SOLVER
!--------------
! End Section
!--------------

!-------------------------------------------
! do not touch the following lines
!-------------------------------------------
#if _NDIMX > 1
#define _DEPTHX 1
#else
#define _DEPTHX 0
#endif
#if _NDIMY > 1
#define _DEPTHY 1
#else
#define _DEPTHY 0
#endif
#if _NDIMZ > 1
#define _DEPTHZ 1
#else
#define _DEPTHZ 0
#endif
#undef _singlePU
#if (_DEPTHX==0) && (_DEPTHY==0) && (_DEPTHZ==0)
#define _singlePU
#endif

!-------------------------------------------
! Section: Set node Lattice size
!-------------------------------------------
#define _NX (_NTX/_NDIMX)
#define _NY (_NTY/_NDIMY)
#define _NZ (_NTZ/_NDIMZ)
!--------------
! End Section
!--------------

!-------------------------------------------
! Section: check block solver
!-------------------------------------------
#ifdef _USE_BLOCK_SOLVER
#ifndef _singlePU
#error Block-solver cannot be used in node-parallel lattice setup. (undef _USE_BLOCK_SOLVER)
#endif
#endif
!--------------
! End Section
!--------------

!!============================================
!! Macros below are used for testing program
!!
!-------------------------------------------
! PARALLEL CHECK 
! (do not define for productive runs!)
!-------------------------------------------
!#define _PARALLELCHECK

!-------------------------------------------
! Reversiblity check
! (do not define for productive runs!)
!-------------------------------------------
!#define _REVCHECK
#undef _REVCHECK

!!============================================
!! Macros below are obsolete.
!!
!!-------------------------------------------
!! Set do loop coalescing depth
!!-------------------------------------------
!!#define _COALESCING 1
!!#define _COALESCING 2
!!#define _COALESCING 3


#endif
