#include <config.h>
module hmc_status_class
!
!= Define HMC algorithm controller class
!
!== Version
!
! $Id: hmc_status_class.F90,v 1.11 2011/06/14 11:10:12 ishikawa Exp $
!
  use comlib
  use lattice_class
  use counter_class
  use hmc_logfile_class
  use hmc_job_manager_class
  use metropolis_test_class
  implicit none
  private

  public :: hmc_status
  public :: new
  public :: delete
  public :: read
  public :: save
  public :: print
  public :: print_log_action
  public :: print_statistics
  public :: is_accepted
  public :: is_do_loop_ended
  public :: is_save_config
  public :: is_hmc_metropolis_on
  public :: get_run_number
  public :: get_trajectory_number
  public :: set_run_number
  public :: set_trajectory_number
  public ::  save_path
  public :: fname_save
  public :: fname_cont

  type :: hmc_status_base
  !
  ! contains HMC controller status
  !
    type(hmc_job_manager),   private :: job_manager
    type(metropolis_status), private :: mp_status      ! HMC Metropolist test status
    type(hmc_logfile),       private :: action_log     ! Action log
    character(CHARLEN),      private :: action_log_fname="Action."
    character(CHARLEN),      private :: hmc_mp_log_fname="dHhist."
  end type

  type(hmc_status_base), save, target :: m_hmc_status
  logical, save :: m_is_initialized = .false.

  type hmc_status
    integer :: idummy(2) = 0
  end type

  interface new
    module procedure new_hmc_status_base
    module procedure new_hmc_status
  end interface

  interface delete
    module procedure delete_hmc_status_base
    module procedure delete_hmc_status
  end interface

  interface read
    module procedure read_hmc_status_base
    module procedure read_hmc_status
  end interface

  interface save
    module procedure save_hmc_status_base
    module procedure save_hmc_status
  end interface

  interface print
    module procedure print_hmc_status_base
    module procedure print_hmc_status
  end interface

  interface print_log_action
    module procedure print_log_action_hmc_status_base
    module procedure print_log_action_hmc_status
  end interface

  interface print_statistics
    module procedure print_stat_hmc_status_base
    module procedure print_stat_hmc_status
  end interface

  interface is_accepted
    module procedure is_accepted_hmc_status_base
    module procedure is_accepted_hmc_status
  end interface

  interface is_do_loop_ended
    module procedure is_do_loop_ended_hmc_status_base
    module procedure is_do_loop_ended_hmc_status
  end interface

  interface is_save_config
    module procedure is_save_config_hmc_status_base
    module procedure is_save_config_hmc_status
  end interface

  interface is_hmc_metropolis_on
    module procedure is_hmc_metropolis_on_hmc_status_base
    module procedure is_hmc_metropolis_on_hmc_status
  end interface

  interface get_run_number
    module procedure get_run_number_hmc_status_base
    module procedure get_run_number_hmc_status
  end interface

  interface get_trajectory_number
    module procedure get_trajectory_number_hmc_status_base
    module procedure get_trajectory_number_hmc_status
  end interface

  interface set_run_number
    module procedure set_run_number_hmc_status_base
    module procedure set_run_number_hmc_status
  end interface

  interface set_trajectory_number
    module procedure set_trajectory_number_hmc_status_base
    module procedure set_trajectory_number_hmc_status
  end interface

contains

subroutine new_hmc_status_base(this)
!
! initialize HMC controller
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  call new(this%job_manager)
  call new(this%action_log,this%action_log_fname)
  call new(this%mp_status,this%hmc_mp_log_fname,    &
 &           mp_test_name="  HMC Metropolis", mp_weight_name="dH")
  return
end subroutine

subroutine delete_hmc_status_base(this)
!
! delete HMC controller
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  call delete(this%action_log)
  call delete(this%mp_status)
  return
end subroutine

subroutine read_hmc_status_base(this,iout)
!
! read HMC controller paramteres from formatted file (by unit number)
!
  implicit none
  type(hmc_status_base), intent(inout) :: this  ! HMC controller
  integer,           intent(in)    :: iout ! file unit number

  call read(this%job_manager,iout)
  call set_hmc_job_manager(this%action_log,this%job_manager)

  return
end subroutine

subroutine save_hmc_status_base(this,iout)
!
! \Save HMC controller status on formated file (by unit number)
!
  implicit none
  type(hmc_status_base), intent(inout) :: this ! HMC controller
  integer,           intent(in)    :: iout ! file unit number

  call save(this%job_manager,iout)

  return
end subroutine

subroutine print_hmc_status_base(this)
!
! \Print out HMC controller status on display
!
  implicit none
  type(hmc_status_base), intent(inout) :: this

  call print(this%job_manager)

  return
end subroutine

subroutine print_stat_hmc_status_base(this)
!
! \Print out HMC controller statistics on display
!
  use lattice_class, only : nodeid
  implicit none
  type(hmc_status_base), intent(inout) :: this
  character(CHARLEN) :: name

  call print_statistics(this%job_manager)
  call print_statistics(this%mp_status)

  return
end subroutine

subroutine print_log_action_hmc_status_base(this,str)
!
! \Print HMC action status to HMC controller
!
! User supply action values in formatted string which is written directly 
! to the action log file.
!
  implicit none
  type(hmc_status_base), intent(inout) :: this  ! HMC controller
  character(len=*),      intent(in)    :: str   ! HMC action strings
  call print(this%action_log,TRIM(str))
  return
end subroutine

function get_run_number_hmc_status_base(this) result(irun)
!
! reutrn HMC run id/ run number
!
  implicit none
  type(hmc_status_base), intent(in) :: this
  integer :: irun
  irun = get_run_number(this%job_manager)
  return
end function

subroutine set_run_number_hmc_status_base(this,irun)
!
! set HMC run id/ run number
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  integer, intent(in) :: irun
  call set_run_number(this%job_manager,irun)
  return
end subroutine

function get_trajectory_number_hmc_status_base(this) result(itraj)
!
! return HMC current trajectory number
!
  implicit none
  type(hmc_status_base), intent(in) :: this
  integer :: itraj
  itraj = get_trajectory_number(this%job_manager)
  return
end function

subroutine set_trajectory_number_hmc_status_base(this,itraj)
!
! set HMC current trajectory number
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  integer, intent(in) :: itraj
  call set_trajectory_number(this%job_manager,itraj)
  return
end subroutine

function is_hmc_metropolis_on_hmc_status_base(this) result(flag)
!
! return HMC Meropolis test switch
!
!  .true.:: Metropolis test is on
! .false.:: Metropolis test is off
!
  implicit none
  type(hmc_status_base), intent(in) :: this
  logical :: flag
  flag = is_hmc_metropolis_on(this%job_manager)
  return
end function

function is_accepted_hmc_status_base(this,dH) result(flag)
!
! return HMC Metropolis test result using action difference value dH
!
!  dH = H1 - H0
!
!  .true.:: accept
! .false.:: reject
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  real(DP), intent(in) :: dH    ! dH=H1-H0
  integer :: itraj
  logical :: flag, onoff

  onoff =  is_hmc_metropolis_on(this%job_manager)
  itraj = get_trajectory_number(this%job_manager)
  flag  = is_accepted(this%mp_status,itraj,dH,onoff)

  return
end function

function is_do_loop_ended_hmc_status_base(this) result(flag)
!
! return HMC job loop status.
!
!  .true.:: HMC run reaches requied sweep. exit loop.
! .false.:: HMC run still remains. go next iteration.
!
  implicit none
  type(hmc_status_base), intent(inout) :: this
  logical :: flag
  flag = is_do_loop_ended(this%job_manager)
  return
end function

function is_save_config_hmc_status_base(this) result(flag)
!
! return whether save or not save configuration for measurment.
!
!  .true.:: user should save configuration for measurement
! .false.:: user need not save configuration for measurement
!
  implicit none
  type(hmc_status_base), intent(in) :: this
  logical :: flag
  flag = is_save_config(this%job_manager)
  return
end function

subroutine new_hmc_status(this)
  implicit none
  class(hmc_status), intent(inout) :: this
  if (.not.m_is_initialized) call new(m_hmc_status)
  m_is_initialized = .true.
  return
end subroutine

subroutine delete_hmc_status(this)
  implicit none
  class(hmc_status), intent(inout) :: this
  if (m_is_initialized) call delete(m_hmc_status)
  m_is_initialized = .false.
  return
end subroutine

subroutine read_hmc_status(this,iout)
  implicit none
  class(hmc_status), intent(inout) :: this  ! HMC controller
  integer, intent(in) :: iout
  call read(m_hmc_status,iout)
  return
end subroutine

subroutine save_hmc_status(this,iout)
  implicit none
  class(hmc_status), intent(inout) :: this ! HMC controller
  integer,           intent(in)    :: iout ! file unit number
  call save(m_hmc_status,iout)
  return
end subroutine

subroutine print_hmc_status(this)
  implicit none
  class(hmc_status), intent(inout) :: this
  call print(m_hmc_status)
  return
end subroutine

subroutine print_stat_hmc_status(this)
  implicit none
  class(hmc_status), intent(inout) :: this
  call print_statistics(m_hmc_status)
  return
end subroutine

subroutine print_log_action_hmc_status(this,str)
  implicit none
  class(hmc_status), intent(inout) :: this  ! HMC controller
  character(len=*),  intent(in)    :: str   ! HMC action strings
  call print_log_action(m_hmc_status,TRIM(str))
  return
end subroutine

function get_run_number_hmc_status(this) result(irun)
  implicit none
  class(hmc_status), intent(in) :: this
  integer :: irun
  irun = get_run_number(m_hmc_status)
  return
end function

subroutine set_run_number_hmc_status(this,irun)
  implicit none
  class(hmc_status), intent(inout) :: this
  integer, intent(in) :: irun
  call set_run_number(m_hmc_status,irun)
  return
end subroutine

function get_trajectory_number_hmc_status(this) result(itraj)
  implicit none
  class(hmc_status), intent(in) :: this
  integer :: itraj
  itraj = get_trajectory_number(m_hmc_status)
  return
end function

subroutine set_trajectory_number_hmc_status(this,itraj)
!
! set HMC current trajectory number
!
  implicit none
  class(hmc_status), intent(inout) :: this
  integer, intent(in) :: itraj
  call set_trajectory_number(m_hmc_status,itraj)
  return
end subroutine

function is_hmc_metropolis_on_hmc_status(this) result(flag)
  implicit none
  class(hmc_status), intent(in) :: this
  logical :: flag
  flag = is_hmc_metropolis_on(m_hmc_status)
  return
end function

function is_accepted_hmc_status(this,dH) result(flag)
  implicit none
  class(hmc_status), intent(inout) :: this
  real(DP), intent(in) :: dH    ! dH=H1-H0
  logical :: flag
  flag  = is_accepted(m_hmc_status,dH)
  return
end function

function is_do_loop_ended_hmc_status(this) result(flag)
  implicit none
  class(hmc_status), intent(inout) :: this
  logical :: flag
  flag = is_do_loop_ended(m_hmc_status)
  return
end function

function is_save_config_hmc_status(this) result(flag)
  implicit none
  class(hmc_status), intent(in) :: this
  logical :: flag
  flag = is_save_config(m_hmc_status)
  return
end function

end module
