module fftw3_class
  use comlib
  use lattice_class
  use quark_prop_class
  implicit none
  private
  public :: fft_qprop,invfft_qprop,new_fft

#include <fftw3.h>
  external dfftw_plan_dft_3d
  external dfftw_execute
  external dfftw_destroy_plan

  type wf_timecolspin_obj
    integer :: itt
    integer :: ic,jc
    integer :: is,js
    integer :: iwf,iwf_nl
    integer :: idummy
  end type
  integer, parameter :: NUM_WF = NTT*COL*COL*SPIN*SPIN

   type(wf_timecolspin_obj), save :: wf_tsc(NUM_WF)                  ! GLOBAL table
   type(wf_timecolspin_obj), save :: wf_tsc_nl(NUM_WF/NPU+1,0:NPU-1) ! node local table
   ! note NUM_WF >= NPU shoud hold.

   integer, save :: NUM_WF_NL   ! node local wavefunc number
   integer, save :: node_site(4,0:NPU-1)

contains

subroutine new_fft
  implicit none
  integer :: iwf,itt,ic,is,jc,js
  integer :: ipu,iwf_nl

  node_site(:,:)=0
  node_site(1,nodeid)=ipsite(1)
  node_site(2,nodeid)=ipsite(2)
  node_site(2,nodeid)=ipsite(3)
  node_site(4,nodeid)=ipsite(4)
  do ipu=0,NPU-1
    call comlib_bcast(node_site(:,ipu),ipu)
  enddo

  ! make (time,col,col,spin,spin) <=> wavefunction index iwf table
  iwf=0
  do js=1,SPIN
  do is=1,SPIN
  do jc=1,COL
  do ic=1,COL
  do itt=1,NTT
    iwf=iwf+1
    wf_tsc(iwf)%itt=itt
    wf_tsc(iwf)%is =is
    wf_tsc(iwf)%js =js
    wf_tsc(iwf)%ic =ic
    wf_tsc(iwf)%jc =jc
    wf_tsc(iwf)%iwf=iwf
    wf_tsc(iwf)%iwf_nl=0 ! not used
    wf_tsc(iwf)%idummy=0 ! not used
  enddo
  enddo
  enddo
  enddo
  enddo

  ! pick up wavefunction number for this MPI process
  do ipu=0,NPU-1
  iwf_nl=0
  NUM_WF_NL=iwf_nl
  do iwf=1,NUM_WF
    if (mod(iwf-1,NPU) == ipu) then
      iwf_nl=iwf_nl+1
      NUM_WF_NL=iwf_nl
      wf_tsc_nl(iwf_nl,ipu) = wf_tsc(iwf)
      wf_tsc_nl(iwf_nl,ipu)%iwf_nl = iwf_nl
    endif
  enddo
  enddo

  return
end subroutine

subroutine fft_qprop(qp)
  implicit none
  type(quark_prop_obj), intent(inout) :: qp
  complex(8), allocatable :: wf(:,:,:,:)
  allocate(wf(NTZ,NTY,NTX,NUM_WF_NL))
  call gather_wavefunc(qp,wf)
  call fft_wavefunc(wf)
  call scatter_wavefunc(wf,qp)
  deallocate(wf)
  return
end subroutine

subroutine invfft_qprop(qp)
  implicit none
  type(quark_prop_obj), intent(inout) :: qp
  complex(8), allocatable :: wf(:,:,:,:)
  allocate(wf(NTZ,NTY,NTX,NUM_WF_NL))
  call gather_wavefunc(qp,wf)
  call invfft_wavefunc(wf)
  call scatter_wavefunc(wf,qp)
  deallocate(wf)
  return
end subroutine

subroutine gather_wavefunc(qp,wf)
  implicit none
  type(quark_prop_obj), intent(in) :: qp
  complex(8), intent(out) :: wf(NTZ,NTY,NTX,NUM_WF_NL)
  integer :: ispt,isp,ipu
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,jc,js,iwf_nl
  integer :: ipux,ipuy,ipuz,iput
  complex(8), allocatable :: tf(:,:),ff(:,:,:,:)

  allocate(tf(NX*NY*NZ*NT,0:NPU-1))
  allocate(ff(NTZ,NTY,NTX,NTT))

  do js=1,SPIN
  do is=1,SPIN
  do jc=1,COL
  do ic=1,COL

    tf(:,:) = (0.0d0,0.0d0)
    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      isp = iz + (iy-1)*NZ + (ix-1)*NZ*NY
      ispt= isp + (it-1)*NZ*NY*NX
      tf(ispt,nodeid) = qp%q(ic,jc,is,js,isp,it)
    enddo
    enddo
    enddo
    enddo

    do ipu=0,NPU-1
      call comlib_bcast(tf(:,ipu),ipu)
    enddo

    do iput=0,NDIMT-1
    do ipux=0,NDIMX-1
    do ipuy=0,NDIMY-1
    do ipuz=0,NDIMZ-1
      do ipu=0,NPU-1
      if ( (node_site(1,ipu) == ipux) .and. &
  &        (node_site(2,ipu) == ipuy) .and. &
  &        (node_site(3,ipu) == ipuz) .and. &
  &        (node_site(4,ipu) == iput) ) then
        do it=1,NT
        do ix=1,NX
        do iy=1,NY
        do iz=1,NZ
          itt = it + iput*NT
          itx = ix + ipux*NX
          ity = iy + ipuy*NY
          itz = iz + ipuz*NZ
          ispt = iz + (iy-1)*NZ + (ix-1)*NZ*NY + (it-1)*NZ*NY*NX
          ff(itz,ity,itx,itt)=tf(ispt,ipu)
        enddo
        enddo
        enddo
        enddo
      endif
      enddo
    enddo
    enddo
    enddo
    enddo

    do itt=1,NTT
      do iwf_nl=1,NUM_WF_NL
      if ( (itt== wf_tsc_nl(iwf_nl,nodeid)%itt) .and. &
  &        (ic == wf_tsc_nl(iwf_nl,nodeid)%ic)  .and. &
  &        (jc == wf_tsc_nl(iwf_nl,nodeid)%jc)  .and. &
  &        (is == wf_tsc_nl(iwf_nl,nodeid)%is)  .and. &
  &        (js == wf_tsc_nl(iwf_nl,nodeid)%js) ) then
        do itx=1,NTX
        do ity=1,NTY
        do itz=1,NTZ
          wf(itz,ity,itx,iwf_nl)=ff(itz,ity,itx,itt)
        enddo
        enddo
        enddo
      endif
      enddo
    enddo

  enddo
  enddo
  enddo
  enddo

  deallocate(tf,ff)
  return
end subroutine

subroutine scatter_wavefunc(wf,qp)
  implicit none
  complex(8), intent(in) :: wf(NTZ,NTY,NTX,NUM_WF_NL)
  type(quark_prop_obj), intent(inout) :: qp
  integer :: ispt,isp,ipu,itsp
  integer :: iz,iy,ix,it,itz,ity,itx,itt
  integer :: ic,is,jc,js,iwf_nl
  integer :: ipux,ipuy,ipuz,iput
  complex(8), allocatable :: tf(:,:)

  allocate(tf(NTX*NTY*NTZ,0:NPU-1))
  do iwf_nl=1,NUM_WF_NL
    tf(:,:) = (0.0d0,0.0d0)
    do itx=1,NTX
    do ity=1,NTY
    do itz=1,NTZ
      itsp = itz + (ity-1)*NTZ + (itx-1)*NTZ*NTY
      tf(itsp,nodeid) = wf(itz,ity,itx,iwf_nl)
    enddo
    enddo
    enddo

    do ipu=0,NPU-1
      call comlib_bcast(tf(:,ipu),ipu)
    enddo

    do ipu=0,NPU-1
      itt = wf_tsc_nl(iwf_nl,ipu)%itt
      ic  = wf_tsc_nl(iwf_nl,ipu)%ic
      jc  = wf_tsc_nl(iwf_nl,ipu)%jc
      is  = wf_tsc_nl(iwf_nl,ipu)%is
      js  = wf_tsc_nl(iwf_nl,ipu)%js
      do itx=1,NTX
      do ity=1,NTY
      do itz=1,NTZ
        ix = mod(itx-1,NX)+1
        iy = mod(ity-1,NY)+1
        iz = mod(itz-1,NZ)+1
        it = mod(itt-1,NT)+1
        ipux = (itx-1)/NX
        ipuy = (ity-1)/NY
        ipuz = (itz-1)/NZ
        iput = (itt-1)/NT
        if ( (ipux == ipsite(1)) .and.  &
  &          (ipuy == ipsite(2)) .and.  &
  &          (ipuz == ipsite(3)) .and.  &
  &          (iput == ipsite(4)) ) then
          isp  = iz + (iy-1)*NZ + (ix-1)*NZ*NY
          itsp = itz + (ity-1)*NTZ + (itx-1)*NTZ*NTY
          qp%q(ic,jc,is,js,isp,it) = tf(itsp,ipu)
        endif
      enddo
      enddo
      enddo
    enddo

  enddo

  deallocate(tf)
  return
end subroutine

subroutine fft_wavefunc(af)
  implicit none
  complex(8), intent(inout) :: af(NTZ,NTY,NTX)
  integer(8) :: plan
  call dfftw_plan_dft_3d(plan,NTZ,NTY,NTX,af,af,FFTW_FORWARD,FFTW_ESTIMATE)
  call dfftw_execute(plan)
  call dfftw_destroy_plan(plan)
  return
end subroutine

subroutine invfft_wavefunc(af)
  implicit none
  complex(8), intent(inout) :: af(NTZ,NTY,NTX)
  integer(8) :: plan
  integer, parameter :: NVOL = (NTZ*NTY*NTX)
  real(8) :: rfac
  integer :: itx,ity,itz
  call dfftw_plan_dft_3d(plan,NTZ,NTY,NTX,af,af,FFTW_BACKWARD,FFTW_ESTIMATE)
  call dfftw_execute(plan)
  call dfftw_destroy_plan(plan)
  rfac = 1.0d0/NVOL
  do itx=1,NTX
  do ity=1,NTY
  do itz=1,NTZ
    af(itz,ity,itx) = af(itz,ity,itx)*rfac
  enddo
  enddo
  enddo
  return
end subroutine

end module
