#include <config.h>

module noisy_wavefunc_class
  use comlib
  use lattice_class
  use wavefunc_class
  implicit none
  private
  public :: wavefunc_param_obj
  public :: LOCAL_WAV
  public ::   EXP_WAV
  public ::  WALL_WAV
  public ::  POLY_WAV
  public :: new,delete,read,get,print,set_noise

  integer, parameter :: NSPACE=NX*NY*NZ

  interface set_noise
    module procedure set_noisy_wavefunc
  end interface

contains

function ispace(ix,iy,iz) result(isp)
  implicit none
  integer :: ix,iy,iz,isp
  isp=iz + (iy-1)*NZ + (ix-1)*NZ*NY
  return
end function

subroutine set_noisy_wavefunc(this)
!==============================================================
! set spin-color diluted complex Z2 noise on wavefunction
!==============================================================
  use random_class
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)
  complex(8), parameter :: z1=(1.0d0,0.0d0)
  integer :: iz,iy,ix,it,ic,is,isp
  complex(8) :: ctmp
  real(8) :: vec(2),cr,ci

  this%wavefunc(:,:,:,:)=z0

  ! set spin-color diluted Z2 noise vector
  do it=1,NT
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    isp=ispace(ix,iy,iz)
    call get_rand3(rand,vec)
    if (vec(1) < 0.5d0) then
      cr=+1.0d0
    else
      cr=-1.0d0
    endif
    if (vec(2) < 0.5d0) then
      ci=+1.0d0
    else
      ci=-1.0d0
    endif
    ctmp=cmplx(cr,ci,kind=KIND(ctmp))/sqrt(2.0d0)
    do is=1,SPIN
    do ic=1,COL
      this%wavefunc(ic,is,isp,it)=ctmp
    enddo
    enddo
  enddo
  enddo
  enddo
  enddo

  return
end subroutine

end module
